local CowTip = CowTip
local CowTip_LuaText = CowTip:NewModule("LuaText", "LibRockEvent-1.0", "LibRockHook-1.0", "LibRockTimer-1.0")

local name, ns = ...
local debug = ns.debug

local ScriptEnv
function CowTip_LuaText:SetScriptEnv(arg)
--	CowTip_LuaText.ScriptEnv = arg
	ScriptEnv = arg
end

--[[
-- Text-provider-code
--]]
local function compileFunction(code)
	local func
	local lua_string = 'return function(unit) if not unit then return \"\" end; '..code..' end'
	local create_func, err = loadstring(lua_string)
	if create_func then
		func = create_func()
		setfenv(func, ScriptEnv)
	else
		geterrorhandler()(err)
	end
	if not func then func = function() return "" end end
	return func
end

local function getOrCreateText(self, code)
	local val = rawget(self, code)
	if not val then
		val = compileFunction(code)
		self[code] = val
	end
	return val
end

local funcs = {
	[""] = function() return "" end,
}
setmetatable(funcs, { __index = getOrCreateText, })

function CowTip_LuaText:RemoveText(code)
	funcs[code] = nil
end

local function autoInitWithTable(self, key)
	local val = rawget(self, key)
	if not val then
		val = {}
		self[key] = val
	end
	return val
end
local failedCalls = setmetatable({}, { __index = autoInitWithTable, })
_G.CowTip.failures = failedCalls

function CowTip_LuaText:GetText(code, unit)
	local func = funcs[code]
	table.insert(failedCalls[code], args)
	local args = { funcs[code](unit) }
	local result = string.format(unpack(args))
	table.remove(failedCalls[code], #failedCalls[code])
	return result
end

local mouseoverUnit = CowTip:GetMouseoverUnit()
--[[
--
--]]

function CowTip_LuaText:OnProfileEnable()
	wipe(funcs)
	for k,v in pairs(self.db.profile.lefts) do
		self:NewText(v, {})
	end
	for k,v in pairs(self.db.profile.rights) do
		self:NewText(v, {})
	end
end

local afk_list
local dnd_list
local offline_list


local L = CowTip:L("CowTip_Text")

local leftGameTooltipStrings = CowTip.leftGameTooltipStrings
local rightGameTooltipStrings = CowTip.rightGameTooltipStrings

local _G = _G
local IsShiftKeyDown = _G.IsShiftKeyDown
local ChatFontNormal = _G.ChatFontNormal
local CreateFrame = _G.CreateFrame
local UIParent = _G.UIParent
local GameTooltip = _G.GameTooltip
local ipairs = _G.ipairs
local GetNumFactions = _G.GetNumFactions
local GetFactionInfo = _G.GetFactionInfo
local UnitPlayerControlled = _G.UnitPlayerControlled
local UnitIsPlayer = _G.UnitIsPlayer
local UnitFactionGroup = _G.UnitFactionGroup
local UnitIsConnected = _G.UnitIsConnected
local UnitIsVisible = _G.UnitIsVisible
local UnitIsPVP = _G.UnitIsPVP
local UnitExists = _G.UnitExists

local kwargs = { unit = 'mouseover' }


function CowTip_LuaText:OnInitialize()
	self.db = CowTip:GetDatabaseNamespace("LuaText")
	CowTip:SetDatabaseNamespaceDefaults("LuaText", "profile", {
		lefts = {
			['*'] = "",
			[1] = [[local name = NameFull(unit)
if not name then
  return UNKNOWN
end
local _, class = UnitClass(unit)
local color = "%1"
if UnitIsPlayer(unit) then
  color = Colorize("%1", ClassColor(unit))
else
  color = Colorize("%1", HostileColor(unit))
end
name = name:gsub(Name(unit), color)
return name
]],
			[2] = [[local guild, rankname, rankindex = GetGuildInfo(unit)
if not guild then
  guild = NPCGuild()
  if not guild then return '' end
  return "<%s>", guild
end
local pguild, _, prankindex = GetGuildInfo("player")
if pguild == guild then
  guild = ("|cff00cc00<%s>|r"):format(guild)
  local rankcolor
  if rankindex > prankindex then
    rankcolor = "|cffff0000%s|r"
  elseif rankindex < prankindex then
    rankcolor = "|cff00ff00%s|r"
  else
    rankcolor = "|cffffff00%s|r"
  end
  rankname = rankcolor:format(rankname)
else
  guild = ("<%s>"):format(guild)
end
return "%s (%d) of %s", rankname, rankindex, guild
]],
			[3] = [[return Faction() or '']],
			[4] = ("return \"%s\""):format(TARGET),
			[5] = [[local level, race = Level(unit), SmartRace(unit)
local class = ""
if UnitIsPlayer(unit) then
  class = Colorize(Class(unit), ClassColor(unit))
else
  class = ""
end
level = Colorize(level, DifficultyColor(unit))
local diff = ShortClassification(unit)
return "Level %s%s %s %s", level, diff, race, class
]],
			[6] = "return 'Status:'",
			[7] = "return 'Zone:'",
			[8] = "return '|T%s:24|t %s', TalentSpecIcon(unit) or '', TalentSpecName(unit) or ''",
		},
		rights = {
			['*'] = "",
			[1] = "",
			[2] = "",
			[3] = "",
			[4] = [[local name = Name(unit .. "target")
if not name then
  return NONE
elseif name == UnitName("player") then
  return Colorize(("<<%s>>"):format(YOU), 0, 0, 255)
else
  return Colorize(name, SmartClassColor(unit .. "target"))
end
]],
			[5] = "",
			[6] = [[return Offline(unit) or AFK(unit) or DND(unit) or Status(unit) or '']],
			[7] = "return Zone() or ''",
			[8] = "return TalentSpec(unit) or ''",
		}
	})
end

function CowTip_LuaText:OnEnable()
	CowTip:GetModule("Text"):ToggleActive(false)
--	ScriptEnv = CowTip_LuaText.ScriptEnv

	self:AddEventListener("UPDATE_FACTION")

	self:AddHook(GameTooltip, "GetUnit", "GameTooltip_GetUnit")
	self:UPDATE_FACTION()
	self:CreateOptions()
	-- gui
end

function CowTip_LuaText:OnDisable()
	wipe(funcs)
end

-- patterns
local LEVEL_start = "^" .. LEVEL
local PVP = _G.PVP
local factionList = {}

local function scanTT(unit)
	local npcguild, faction, zone
	-- scan for faction
	if not unit then
		wipe(ScriptEnv.orig)
		return
	end

	local left_2 = leftGameTooltipStrings[2]:GetText() or ""
	local hasGuild = not left_2:find(LEVEL_start)
	if hasGuild and not UnitIsPlayer(unit) then
		debug("npcguild found ", left_2)
		npcguild = left_2
	end
	local left_3 = leftGameTooltipStrings[3]:GetText()
	local left_4 = leftGameTooltipStrings[4]:GetText()
	local left_5 = leftGameTooltipStrings[5]:GetText()
	local faction
	if not UnitPlayerControlled(unit) and not UnitIsPlayer(unit) then
		faction = hasGuild and left_4 or left_3
		if faction == PVP or not factionList[faction] then
			faction = nil
		end
	end
	-- scan for zone
	local zone
	if not UnitIsVisible(unit) and UnitIsConnected(unit) then
		if hasGuild and faction then
			zone = left_5
		elseif hasGuild or faction then
			zone = left_4
		else
			zone = left_3
		end
	end
	wipe(ScriptEnv.orig)
	ScriptEnv.orig.faction = faction
	ScriptEnv.orig.zone = zone
	ScriptEnv.orig.npcguild = npcguild
end

function CowTip_LuaText:UPDATE_FACTION()
	for i = 1, GetNumFactions() do
		local name = GetFactionInfo(i)
		factionList[name] = true
	end
end

--[[
How normal tooltips look:
{Name}
{[Guild] : ~~~}
Level {Num} {[(Classification)]}
{NPC? [Faction] : ~~~}
{Offline? Offline : !IsVisible? [Zone] : IsPVP? [PvP] : ~~~}
]]

local lastNumRegistered = 0
local linesToAdd = {}
local linesToAddR = {}
local linesToAddG = {}
local linesToAddB = {}
local linesToAddRight = {}
local linesToAddRightR = {}
local linesToAddRightG = {}
local linesToAddRightB = {}

local lastUnitName, lastUnit

function CowTip_LuaText:RefixLines(lastLine)
	lastNumRegistered = 0
	local LinesAdded = 0

	for i = lastLine, GameTooltip:NumLines() do
		local left = leftGameTooltipStrings[i]
		local j = i - lastLine + 1
		linesToAdd[j] = left:GetText()
		local r, g, b = left:GetTextColor()
		linesToAddR[j] = r
		linesToAddG[j] = g
		linesToAddB[j] = b
		local right = rightGameTooltipStrings[i]
		if right:IsShown() then
			linesToAddRight[j] = right:GetText()
			local r, g, b = right:GetTextColor()
			linesToAddRightR[j] = r
			linesToAddRightG[j] = g
			linesToAddRightB[j] = b
		end
		LinesAdded = LinesAdded + 1
	end
	GameTooltip:ClearLines()
	local db = self.db.profile
	local lefts, rights = db.lefts, db.rights
	local num = 0
	local unit = CowTip:GetMouseoverUnit()
	for i = 1, 30 do
		local left, right = lefts[i], rights[i]
		local left_ev, right_ev
		if left and left:trim() == "" then
			left = nil
		else
			--left_ev = tostring(DogTag:Evaluate(left, "Unit", kwargs) or '')
			left_ev = tostring(self:GetText(left, unit))
			if left_ev:trim() == "" then
				left_ev = nil
			end
		end
		if right and right:trim() == "" then
			right = nil
		else
--			right_ev = tostring(DogTag:Evaluate(right, "Unit", kwargs) or '')
			right_ev = tostring(self:GetText(right, unit))
			if right_ev:trim() == "" then
				right_ev = nil
			end
		end
		if left_ev or right_ev then
			if right then
				--if right_ev or (left and left_ev and left:find("%[")) then
				if right_ev or db.alwaysShowLefts then
					num = num + 1
					GameTooltip:AddDoubleLine(left_ev or ' ', right_ev, 1, 1, 1, 1, 1, 1)
					local leftFS = leftGameTooltipStrings[num]
					local rightFS = rightGameTooltipStrings[num]
	--				DogTag:AddFontString(leftFS, GameTooltip, left or ' ', "Unit", kwargs)
	--				DogTag:AddFontString(rightFS, GameTooltip, right, "Unit", kwargs)
				end
			elseif left then
				if left_ev then
					num = num + 1
					GameTooltip:AddLine(left_ev, 1, 1, 1, true)
					local leftFS = leftGameTooltipStrings[num]
					leftFS:SetTextColor(1, 1, 1)
	--				DogTag:AddFontString(leftFS, GameTooltip, left, "Unit", kwargs)
				end
			end
		end
	end
	if lastNumRegistered < num then
		lastNumRegistered = num
	end

	for i = 1, LinesAdded do
		local left = linesToAdd[i]
		local right = linesToAddRight[i]
		if right then
			GameTooltip:AddDoubleLine(left, right, linesToAddR[i], linesToAddG[i], linesToAddB[i], linesToAddRightR[i], linesToAddRightG[i], linesToAddRightB[i])
		else
			GameTooltip:AddLine(left, linesToAddR[i], linesToAddG[i], linesToAddB[i], true)
		end
		linesToAdd[i] = nil
		linesToAddR[i] = nil
		linesToAddG[i] = nil
		linesToAddB[i] = nil
		linesToAddRight[i] = nil
		linesToAddRightR[i] = nil
		linesToAddRightG[i] = nil
		linesToAddRightB[i] = nil
	end
end

function CowTip_LuaText:GameTooltip_GetUnit(this)
	if not lastUnitName then
		return self.hooks[GameTooltip].GetUnit(GameTooltip)
	end
	return lastUnitName, lastUnit
end

local function delayedSetUnit()
	local unit = select(2, GameTooltip:GetUnit())-- CowTip:GetMouseoverUnit()
	debug("delayed set unit ", unit)
	if unit and UnitExists(unit) then
		CowTip_LuaText:OnTooltipSetUnit2()
		GameTooltip:Show()
	end
end

function CowTip_LuaText:OnTooltipSetUnit()
	local unit = select(2, GameTooltip:GetUnit())-- CowTip:GetMouseoverUnit()
	if unit and UnitExists(unit) then
		debug("set unit ", unit, " immediately")
		self:OnTooltipSetUnit2()
		self:RemoveTimer("CowTip_LuaText_SetLater")
	elseif not self:HasTimer("CowTip_LuaText_SetLater") then
		self:AddTimer("CowTip_LuaText_SetLater", 0.1, delayedSetUnit)
	end
end

function CowTip_LuaText:OnTooltipSetUnit2()
	lastUnitName, lastUnit = self.hooks[GameTooltip].GetUnit(GameTooltip)
	local num = 2
	local text_2 = leftGameTooltipStrings[2]:GetText()
	if not text_2 then
		num = num - 1
	elseif not text_2:find(LEVEL_start) then
		-- has a guild
		num = num + 1
	end
	local mouseover_unit = CowTip:GetMouseoverUnit()
	if not UnitPlayerControlled(mouseover_unit) and not UnitIsPlayer(mouseover_unit) then
		-- npc, faction line
		local factionText = leftGameTooltipStrings[num+1]:GetText()
		if factionText == PVP then
			factionText = nil
		end
		if factionText and (factionList[factionText] or UnitFactionGroup(mouseover_unit)) then
			num = num + 1
		end
	end
	if not UnitIsConnected(mouseover_unit) or not UnitIsVisible(mouseover_unit) or UnitIsPVP(mouseover_unit) then
		-- offline, zone, or pvp
		num = num + 1
	end
	scanTT(mouseover_unit)
	self:RefixLines(num+1)
	self:AddRepeatingTimer("CowTip_LuaText_UpdateTexts", 1, "UpdateTexts")
end

function CowTip_LuaText:UpdateTexts()
	if not lastUnit or not UnitExists(mouseover_unit) then
		return
	end
	self:RefixLines(lastNumRegistered+1)

	GameTooltip:Show()
end

function CowTip_LuaText:OnTooltipHide()
	for i = 1, lastNumRegistered do
--		DogTag:RemoveFontString(leftGameTooltipStrings[i])
--		DogTag:RemoveFontString(rightGameTooltipStrings[i])
	end
	lastNumRegistered = 0
	lastUnitName, lastUnit = nil, nil

	self:RemoveTimer("CowTip_LuaText_UpdateTexts")
end

local function CleanCode(text)
--	text = DogTag:CleanCode(text)
	text = text:gsub("%s%s*", " ")
	return text
end

local function fixColor()
	GameTooltipTextLeft1:SetTextColor(1, 1, 1)
end

function CowTip_LuaText:PostTooltipSetUnit()
	self:AddTimer(0, fixColor)
end
